#!/usr/bin/perl -w

# BETAWRAP
# Created by
# Joseph Gagliano
# Anshul Kundaje
# Deniz Sarioz
# Daita Domnica Nicolae

use POSIX;

#Parameters for the gap penalty
$P = 1;  
$Q = 10;
$mean_native_gap = 12;

# Score matrices
$aastr = "ACDEFGHIKLMNPQRSTVWY";   # Indices of the score matrix
@burmat = getMatrix("bur.dat");    # Score matrix for buried residues
@expmat = getMatrix("exp.dat");    # Score matrix for exposed residues

$t_len = 9;                        # Length of the B2-T2-B3 segment
###############################################################################
#Checking for arguments

if ($#ARGV != 0){
    die "Usage:	 betasplice <input>\n";
}

################################################################################
#Read in the  Data

$testFile = $ARGV[0];              # FASTA file of the protein to be tested
if (! -r $testFile){
    die "Can't open the Input Data File";
}

open(INPUT, "$testFile") || die "Error Opening Input File $testFile $!\n";
@testData = <INPUT>;
close(INPUT);

################################################################################
# Processing of file to remove unwanted lines and concatenated the whole protein sequence
# into one string

@seq_array = ();                    # The linear protein sequence

foreach $line (@testData){
    if( $line =~ m/>/ ){}           # Eliminate header line
    else{
	
	$line =~ s/([0-$t_len])//g;       # Strip out the numbers
	$line =~ s/\s//g;                 # Strip out the spaces, tabs and new lines
	push(@seq_array, $line);                  
    }
}

$seq=join('', @seq_array);         # Concatenate each line into a single string
print "$testFile\n";
print $seq."\n";
print "=================================\n";
#print "\n\n";
#print length($seq);
#print "\n";

###################################################################################
# Finding Potential Rungs

@rungs=();                       # Each element of the array is a string that contains the potential rungs

# Find all possible rungs by using a silding window of size 9 and match it with the template
for($i=0; $i<= (length($seq)-$t_len) ; $i++){
    $s =  substr($seq,$i, $t_len);
    if( $s =~ m/[AFILMVWY].[AFILMVWY]..[^DERK].[AFILMVWYG]./oi ){
	push(@rungs, $s);
	push(@rungpos, $i);
#	print "$s\n";
    }
}

###################################################################################
#Create a matrix containing the 'Rung-segment alignment Score' for every
#possible pair of rung-segments we found in the above step

$num_rungs = $#rungs+1;

print ("NUMRUNGS:  $num_rungs\n");

if ($num_rungs < 5) {     # If number of rungs < 5 then no wraps are possible and protein is rejected
    print "---------------------No Valid Rungs: THE PROTEIN $testFile IS REJECTED\n";
    exit;
}

for($i=0; $i<$num_rungs; $i++){
    for($j=0; $j<$num_rungs; $j++){
	if ($i != $j) {
	    $align_matrix[$i][$j] = alignScore($rungs[$i],$rungs[$j], $rungpos[$i], $rungpos[$j]);
	} else {
	    $align_matrix[$i][$j] = -1000000000;     # Self-alignment score is forced to be very small
	}
    }
}

#for($i=0; $i<$num_rungs; $i++){
#    for($j=0; $j<$num_rungs; $j++){
#	
#	printf("%.2f  ", $align_matrix[$i][$j]);
	
#    }
#    print "\n";
#}

##############################################################################################
# Create 2 matrices that hold the scores and positional indices of top five candidates for next rungs 
# aligning to each rung in the 

for($j=0; $j<$num_rungs; $j++){         # For each column in rung-segment alignement score matrix
    @t=();
    
    for ($i=0; $i<$num_rungs; $i++) {
 	$t[$i] = $align_matrix[$i][$j];
    }
    
    @t5rungs = topFive(@t);   # returns a 2D array: column1- scores in ascending order
                              # column2-corresponding IDs of the scores in column 1

# Push the top 5 aligned scores and IDs for each rung-segment into a common array 
# in descending order of scores
    for ($x=0; $x<5; $x++){
	#printf("%.2f",  $t5rungs[$num_rungs-1-$x][0]);
	#print " $t5rungs[$num_rungs-1-$x][1]\n";
	$top5rungScore[$j][$x] = $t5rungs[$num_rungs-1-$x][0];   #descending order
	$top5rungID[$j][$x] = $t5rungs[$num_rungs-1-$x][1];      #corresponding IDs

    }
}

#for ($i=0; $i<$num_rungs; $i++){
    #print "$i\t| ";
 #   for($j=0; $j<5; $j++){
	#printf ("%d\t", $top5rungID[$i][$j]);
    #}
    #print "\n";
    
#}
#for ($i=0; $i<$num_rungs; $i++){
 #   print "$i\t| ";
  #  for($j=0; $j<5; $j++){
	#printf ("%.2f\t", $top5rungScore[$i][$j]);
	
#    }
 #   print "\n";
    
#}

#############################################################################################
# Find all wraps in forward direction using wrap generator algorithm using depth-first search

$threshold= -80;                          # (4hops*20)threshold to filter out wraps with very poor scores
$score=0;                                       # score variable per wrap
#print "\nCalculating scores forward:\n\n";

for $n (0..$num_rungs-5) {   # Donot run it for last 4 rung segments as we cannot have a wrap of < 5 rungs
    $nodes[0]=$n;
    
    # The four for loops represent 4 hops 
    for $i (0..4) {
	$nodes[1]=$top5rungID[$nodes[0]][$i];
	next if ($nodes[0]>=$nodes[1]);          # check for ascending order of indices
	$score += $top5rungScore[$nodes[0]][$i]; # if condition is satisfied add score of alignment
	
	for $j (0..4) {
	    $nodes[2]=$top5rungID[$nodes[1]][$j];
	    next if ($nodes[1]>=$nodes[2]);
	    $score += $top5rungScore[$nodes[1]][$j];
		
	    for $k (0..4) {
		$nodes[3]=$top5rungID[$nodes[2]][$k];
		next if ($nodes[2]>=$nodes[3]);
		$score += $top5rungScore[$nodes[2]][$k];
		
		for $m (0..4) {
		    $nodes[4]=$top5rungID[$nodes[3]][$m];
		    next if ($nodes[3]>=$nodes[4]);
		    $score += $top5rungScore[$nodes[3]][$m];
		    if ($score>$threshold) {
			#print" ";
			@t = ();
			foreach $node (@nodes) {
			    #print "$node ";
			    push @t, $node;               # Push all nodes of the wrap into an array
			}
			$scoreToReport = $score/4;
			#print " score: $scoreToReport\n";
			push @t, $scoreToReport;          # Push score into the wrap array
			push @wraps, [ @t ];              # Push each wrap array into a array of all wraps
		    }
		    $score -= $top5rungScore[$nodes[3]][$m]; # go one hop up the tree from the leaf node
		}
		$score -= $top5rungScore[$nodes[2]][$k]; # move upto to the 2nd hop from leaf node
	    }
	    $score -= $top5rungScore[$nodes[1]][$j];  # move upto the 3rd hop from leaf node
	}
	$score -= $top5rungScore[$nodes[0]][$i];   # move upto the 4th hop from leaf node
    }
}

# Find all wraps in the backward direction using wrap generator algorithm using depth-first search

#print "\nCalculating scores backward:\n\n";
for ($n=$num_rungs-1; $n>=4; $n--) {   # Donot run it for first 4 as we cannot have a wrap of < 5 rungs
    $nodes[0]=$n;
    
    for $i (0..4) {
	$nodes[1]=$top5rungID[$nodes[0]][$i];
	next if ($nodes[0]<=$nodes[1]);  # check for descending order   
	$score += $top5rungScore[$nodes[0]][$i];
	
	for $j (0..4) {
	    $nodes[2]=$top5rungID[$nodes[1]][$j];
	    next if ($nodes[1]<=$nodes[2]);
	    $score += $top5rungScore[$nodes[1]][$j];
	    
	    for $k (0..4) {
		$nodes[3]=$top5rungID[$nodes[2]][$k];
		next if ($nodes[2]<=$nodes[3]);
		$score += $top5rungScore[$nodes[2]][$k];

		for $m (0..4) {
		    $nodes[4]=$top5rungID[$nodes[3]][$m];
		    next if ($nodes[3]<=$nodes[4]);
		    $score += $top5rungScore[$nodes[3]][$m];
		    if ($score>$threshold) {
			#print" ";
			@t = ();
			foreach $node (@nodes) {
			 #   print "$node ";
			    push @t, $node;
			}
			$scoreToReport = $score/4;
			#print " score: $scoreToReport\n";
			push @t, $scoreToReport;
			push @wraps, [ @t ];
		    }
		    $score -= $top5rungScore[$nodes[3]][$m];
		}
		$score -= $top5rungScore[$nodes[2]][$k];
	    }
	    $score -= $top5rungScore[$nodes[1]][$j];
	}
	$score -= $top5rungScore[$nodes[0]][$i];
    }
}

# sort @wraps descending by score i.e. 6th cell of each row of the array which holds the score
# ..hence $b->[5] and $a->[5]

@wraps = sort {$b->[5] <=> $a->[5]} @wraps;

print "A TOTAL OF $#wraps WRAPS WERE FOUND\n";

# Reject protein if less than 10 potential wraps are found
if ($#wraps < 9){
    print "-------------------THE PROTEIN $testFile IS REJECTED-less than 10 wraps\n";
    exit;
}

#print "PRINTING VALID WRAPS\n=====================\n";
@valid_wraps = ();           # Flag array for each wrap: 1=valid wrap; 0=invalid wrap

for ($i=0; $i<=$#wraps; $i++){  # initialise all wraps as valid
    $valid_wraps[$i] = 1;
}


##################################################################################################
#perform Filtering of T3-B1-T1 section of wraps
filterT1B1();

print "FILTERING SEQUENCES\n";
filterT1B1_2();

#exit;

##################################################################################################
# Perform Filtering of the hydrophobics at T2

$i=0;

foreach $wrap (@wraps){
    @t=@$wrap;
    if(! filterHydrophob(@t) ){   # returns flag value for the wrap
	$valid_wraps[$i] = 0;
    }
 #   print "\n";    
    $i++;
}
##################################################################################################
#perform Filtering of alpha helices for all of the wraps
filterAlpha();

##################################################################################################
$i=0;
#print "\nPrinting EVERYTHING:\n";
#foreach $wrap (@wraps){
 #   @t=@$wrap;
  #  foreach $tval (@t) {
	#print "$tval ";
   # }
   # print "$valid_wraps[$i]\n";
   # $i++;
    
#}
#print "\n============\n";
print "Valid Wraps\n";

$i=25;
$x=0;
$numvalwraps = 0;
print "The B2 positions\n";
foreach $wrap (@wraps){
    last if ($i==0);
    if( $valid_wraps[$x] == 1){
	$numvalwraps++;
	@t=@$wrap;
	$score = pop(@t);
	foreach $tval (@t) {
	    $outval = 1 + $rungpos[$tval]; # fudge index to start with 1
	    print "$outval ";
	}
	print " $score\n ";    
#	print "$valid_wraps[$x]\n";
	$i--;
    }
    $x++;
}
print "\n============\n";
if ($numvalwraps < 12){
    print " ---------------------PROTEIN $testFile IS REJECTED AS A BETA HELIX\n";
    print "==============\n";
    exit;
}
else{
    print " ++++++++++++++++++++++++++PROTEIN $testFile HAS A BETA HELIX\n";
    print "==============\n";
    exit;
}

#################################################################################################
#################################################################################################
# SUBROUTINES
#################################################################################################
#################################################################################################
# Takes a set of scores and sorts them in ascending order. 
# Returns a 2D array: Column1: Sorted scores
#                     Column2: Corresponding indices

sub topFive{
    @at = @_;
    @bt = ();
    
    # creates a 2D array with scores and indices
    for ($i=0; $i<=$#at; $i++){
	@temp = ($at[$i], $i);
	$bt[$i] = [@temp];
    }
    
    # sorting based on first column i.e. scores; the 2nd column i.e. indices are shifted accordingly
    @b_sorted = sort {$a->[0] <=> $b->[0]}@bt;
    return (@b_sorted);
}

##############################################################################################
# Computes the alignment score of two sequences based on the scoring matrices

sub alignScore {
    # calculate probability of rung a given rung b
    my $a = $_[0];  # two params are sequences.
    my $b = $_[1];  # we align segment b to a
    my $pa = $_[2]; # position of first residue of first sequence
    my $pb = $_[3]; # position of first residue of second sequence
    my $c = "";
    my ($i, $ares, $bres, $mat_det, $p, $sumlogprobs);

    if ($_[4]) { $c = $_[4]; }  #the pattern to be aligned, defaults to bxbx--xbx

# The sequences should both be B2-T2-B3 candidates.
    if ((length($a) != 9) || (length($b) != 9)) {
	die "Rungs MUST be of length 9 in alignScore().\n";
    }
#    unless ($pa && $pb) {
#	die "alignScore did not receive positions for $a and $b.  bye\n";
#    }
    if ((length($c) != 9) || (! ($c =~ /\A(x|b|-)+\Z/) )) {
#	print "alignScore() is using default buried-exposed pattern.\n";
	$c="bxbx--xbx";                            # default template if undefined or bad
    }

    $a =~ tr/[a-z]/[A-Z]/; #make sure sequences are uppercase
    $b =~ tr/[a-z]/[A-Z]/;
 #   print "a: $a, b: $b, c: $c\n";
    my @A = split(//, $a);
    my @B = split(//, $b);
    my @C = split(//, $c);
    $sumlogprobs = 0.0;

# Defined global variables
# $aastr = "ACDEFGHIKLMNPQRSTVWY";   Indices of the score matrix
# @burmat = getMatrix("bur.dat");    Score matrix for buried residues
# @expmat = getMatrix("exp.dat");    Score matrix for exposed residues

    for ($i=0; $i<9; $i++) {
	$mat_det = $C[$i];
	next if ($mat_det eq "-");
	$ares = $A[$i];
	$aastr =~ /$ares/g;
	$ares = pos($aastr)-1;
	pos($aastr) = 0;
	$bres = $B[$i];
	$aastr =~ /$bres/g;
	$bres = pos($aastr)-1;
	pos($aastr) = 0;
	if ($mat_det eq "x") {
	    #      push @rawprobs, $expmat[$ares][$bres];
	    $sumlogprobs += 0.5 * log $expmat[$ares][$bres];
	    #      $sumweights += 0.5;
	} elsif ($mat_det eq "b") {
	    #      push @rawprobs, $burmat[$ares][$bres];
	    $sumlogprobs += log $burmat[$ares][$bres];
	    #      $sumweights += 1.0;
	} else { # for debugging
	    die "Died in alignScore() due to very weird matrix determiner/\n";
	}
    }
# Calculate penalty and subtract from alignment score
    my $penalty = $P * floor(abs ( (abs($pa-$pb)-9) - $mean_native_gap)/$Q);
    return $sumlogprobs-$penalty;
}

###########################################################################################
# Gets the inward and outward matrices in the desired form

sub getMatrix {
    # divides by 100 to get probability for each entry
    unless ($_[0]) {
	die "Have to call getMatrix() with file name\n";
    }
    my $matrixfile = $_[0];
    my ($i, $j, $m);
    my (@temp, @tmp);
    my @A = ();
    open (MATR, $matrixfile) ||
	die "Can not open $matrixfile for reading in getMatrix().\n";
    while (<MATR>) {
	next if /^\s*\Z/;
	@temp = split;
	@tmp = ();
	foreach $i (@temp) {
	    push @tmp, $i/100;
	}
	push @A, [ @tmp ];
    }
    return @A;
}

##################################################################################################
# Hydrophobic Filter for T2

sub filterHydrophob{
    @a = @_;
    $count = 0;
    pop(@a);

    foreach $x (@a){
#	print $rungs[$x],"  ";
	@r = split(//,$rungs[$x]);
#	print $r[4],"\n";
	if ( $r[4] =~ /[VILFMW]/i){
	    $count++;
	    if ($count >1){
#		print "INVALID\n";	
		return 0;
		
	    }
	}
    }
 #   print "VALID\n";
    return 1;
    
}
######################################################################################################
# The Alpha helix filter
sub filterAlpha{
    
    # Please change path of java bin directory as required
    $a = `/usr/java/bin/java -classpath ./pred2ary.jar pred2ary -p ./$testFile -op /dev/stdout`;

    @lines = split('\n',$a);
    $ignore = 1;
    
    foreach $line (@lines){
	if($line =~ m/expected accuracy/) {
	    $ignore = 0;
	    next;
	}
	next if ($ignore);
	last if ($line =~ m/^\s*\Z|Doing class prediction/);
	
	push (@predset, $line);
	
    }
    
    foreach $line (@predset){
	@l2 = split(/\s+/, $line);
	#print $line,"  ";
	foreach $x (@l2){
	}
	#print "\n";
	if($l2[4] =~ /H/i){
	    push(@alphapos, $l2[1]);
	}
    }
    
    #Go through each rung to see if it has at least 4
    #residues that are at alpha helix positions to be rejected
    
    for ($i=0; $i<=$#rungpos; $i++){
	$rungalpha[$i]=0;
    }
    
    $i=0;
    foreach $n (@rungpos){
	$rpos = $n+1;  #each rpos is offset by 1
	foreach $x (@alphapos){
	    if (($x >= $rpos) && ($x <= $rpos+8)){
		#print "INCREMENTED rungpos $rpos for $x\n";
		$rungalpha[$i]++;
	    }
	    
	}
	$i++;
	
    }
    
#    foreach $r (@rungalpha){
#	print "$r\n";
	
 #   }
    
    #Now check each wrap for a rung that is classified as an alpha helix
    #Make Threshold 5 since Pred2ary tends to overestimate alpha helix length
    
    my $alpha_thresh = 5;

    $w=0; #index for wraps
    foreach $wrap (@wraps){
	@t=@$wrap;
	pop(@t);
	foreach $r (@t){  #for each rung index in the wrap
	    if ($rungalpha[$r] > $alpha_thresh){
		$valid_wraps[$w] = 0;
	    }
	}
	$w++;
    }
}

#############################################################################################
# The T1-B1 filter Part 1

sub filterT1B1{

    # Check the length of T3-B1-T1 to verify they are at least min_length_thresh long
    # but less than max_length_tresh

    my $min_length_thresh =17;
    my $max_length_thresh =55;

    $w=0;
    foreach $wrap (@wraps){
	@t=@$wrap;
	pop(@t);
	for ($i=0; $i<4; $i++){
	    $len = abs($rungpos[$t[$i]] - $rungpos[$t[$i+1]]) ;
	    if( $len > $max_length_thresh || $len < $min_length_thresh){
		$valid_wraps[$w] = 0;
	    }
	}
	$w++;	
    }    
}

################################################################################################
# The T1-B1 Filter Part 2

sub filterT1B1_2{
    $index=0;
    foreach $wrap (@wraps){            # for each wrap
	next if ($valid_wraps[$index++] < 1);    # avoid wraps that are already rejected
	@segments = ();             # holds the T3-B1-T1 segments
	@t=@$wrap;
	my $score = pop(@t);        
	my $s;
	my $offset;                # length of the T3-B1-T1 segment
	for ($i=0; $i<4; $i++){    # find the segments and push them into the segment array
	    my $x = $rungpos[$t[$i]];
	    if ($rungpos[$t[$i]] < $rungpos[$t[$i+1]] ){
		$offset = abs( ($rungpos[$t[$i]]+9) - ($rungpos[$t[$i+1]]));
		$s = substr($seq, $rungpos[$t[$i]]+9+4, $offset-3);  # restrict search to first_rung_end+4 
	                                                             # residues and next_rung_beginning-3
		                                                     # residues to account for T3 and T1 resp
	    }
	    else {
		$offset = abs( ($rungpos[$t[$i+1]]+9) - ($rungpos[$t[$i]]));
		$s = substr($seq, $rungpos[$t[$i+1]]+9+4, $offset-3);
	    }
	    push (@segments, $s);
	}
	
#	foreach $s3 (@segments){
#	    print"$s3\n";
#	}
	
	#Apply the Phi_X_Phi_X template to the first segment found
	foreach $s1 (@segments){
	    @matches = ($s1=~ /[AFILMVWYG].[AFILMVWYG]./gi);  # push all matches into the matches array
	    if($#matches <0) {           # if no matches reject the whole wrap
		$valid_wraps[$index]=0;
		last;
	    }
	    #    print "------------------\n";
	    # foreach $m (@matches){
	    #	print "$m\n";
	    #   }
	}
#	print "\n";
	
    }
}
#########################################################################################################
















